/*
 * Copyright (c) 2018-2019, Texas Instruments Incorporated
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * *  Redistributions of source code must retain the above copyright
 *     notice, this list of conditions and the following disclaimer.
 *
 * *  Redistributions in binary form must reproduce the above copyright
 *     notice, this list of conditions and the following disclaimer in the
 *     documentation and/or other materials provided with the distribution.
 *
 * *  Neither the name of Texas Instruments Incorporated nor the names of
 *     its contributors may be used to endorse or promote products derived
 *     from this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO,
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR
 * CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
 * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
 * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
 * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
 * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
 * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE,
 * EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/*
 *  ======== spimaster.c ========
 */
#include <stddef.h>
#include <stdint.h>
#include <string.h>

/* POSIX Header files */
#include <pthread.h>
#include <semaphore.h>
#include <unistd.h>

/* Driver Header files */
#include <ti/drivers/GPIO.h>
#include <ti/drivers/SPI.h>
#include <ti/display/Display.h>

/* Example/Board Header files */
#include "Board.h"

#define THREADSTACKSIZE (1024)

#define SPI_MSG_LENGTH  (1)
uint16_t MASTER_MSG[4]= {0x0000,0x0800,0x1000,0x1800};

#define MAX_LOOP        (10)

SPI_Handle      masterSpi;
SPI_Params      spiParams;
SPI_Transaction transaction;
bool            transferOK;
int32_t         status;
uint16_t        RX_Buffer_1,RX_Buffer_2,RX_Buffer_3,RX_Buffer_4,i,j;

static Display_Handle display;

uint16_t masterRxBuffer[1];
uint16_t masterTxBuffer[1];

/* Semaphore to block master until slave is ready for transfer */
sem_t masterSem;

void read_adc_channels();

/*
 *  ======== slaveReadyFxn ========
 *  Callback function for the GPIO interrupt on Board_SPI_SLAVE_READY.
 */
//void slaveReadyFxn(uint_least8_t index)
//{
//    sem_post(&masterSem);
//}

/*
 *  ======== masterThread ========
 *  Master SPI sends a message to slave while simultaneously receiving a
 *  message from the slave.
 */
void *masterThread(void *arg0)
{
    /*
     * Board_SPI_MASTER_READY & Board_SPI_SLAVE_READY are GPIO pins connected
     * between the master & slave.  These pins are used to synchronize
     * the master & slave applications via a small 'handshake'.  The pins
     * are later used to synchronize transfers & ensure the master will not
     * start a transfer until the slave is ready.  These pins behave
     * differently between spimaster & spislave examples:
     *
     * spimaster example:
     *     * Board_SPI_MASTER_READY is configured as an output pin.  During the
     *       'handshake' this pin is changed from low to high output.  This
     *       notifies the slave the master is ready to run the application.
     *       Afterwards, the pin is used by the master to notify the slave it
     *       has opened Board_SPI_MASTER.  When Board_SPI_MASTER is opened, this
     *       pin will be pulled low.
     *
     *     * Board_SPI_SLAVE_READY is configured as an input pin. During the
     *       'handshake' this pin is read & a high value will indicate the slave
     *       ready to run the application.  Afterwards, a falling edge interrupt
     *       will be configured on this pin.  When the slave is ready to perform
     *       a transfer, it will pull this pin low.
     *
     * Below we set Board_SPI_MASTER_READY & Board_SPI_SLAVE_READY initial
     * conditions for the 'handshake'.
     */
//    GPIO_setConfig(Board_SPI_MASTER_READY, GPIO_CFG_OUTPUT | GPIO_CFG_OUT_LOW);
//    GPIO_setConfig(Board_SPI_SLAVE_READY, GPIO_CFG_INPUT);

    /*
     * Handshake - Set Board_SPI_MASTER_READY high to indicate master is ready
     * to run.  Wait Board_SPI_SLAVE_READY to be high.
     */
    //GPIO_write(Board_SPI_MASTER_READY, 1);
    //while (GPIO_read(Board_SPI_SLAVE_READY) == 0) {}

    /* Handshake complete; now configure interrupt on Board_SPI_SLAVE_READY */
//    GPIO_setConfig(Board_SPI_SLAVE_READY, GPIO_CFG_IN_PU | GPIO_CFG_IN_INT_FALLING);
//    GPIO_setCallback(Board_SPI_SLAVE_READY, slaveReadyFxn);
//    GPIO_enableInt(Board_SPI_SLAVE_READY);

    /*
     * Create synchronization semaphore; the master will wait on this semaphore
     * until the slave is ready.
     */
//    status = sem_init(&masterSem, 0, 0);
//    if (status != 0) {
//        Display_printf(display, 0, 0, "Error creating masterSem\n");
//
//        while(1);
//    }

    /* Open SPI as master (default) */
    SPI_Params_init(&spiParams);
    spiParams.frameFormat = SPI_POL0_PHA0;
    spiParams.dataSize = 16;
    spiParams.bitRate = 100000;
    //masterSpi = SPI_open(Board_SPI_MASTER, &spiParams);
    Display_printf(display, 0, 0, "Master SPI initialized\n");

    /*
     * Master has opened Board_SPI_MASTER; set Board_SPI_MASTER_READY high to
     * inform the slave.
     */
    //GPIO_write(Board_SPI_MASTER_READY, 0);

    while(1){
        read_adc_channels();
        usleep(500000);
    }

    /* Example complete - set pins to a known state */
    //GPIO_disableInt(Board_SPI_SLAVE_READY);
    //GPIO_setConfig(Board_SPI_SLAVE_READY, GPIO_CFG_OUTPUT | GPIO_CFG_OUT_LOW);
    //GPIO_write(Board_SPI_MASTER_READY, 0);
}

/*
 *  ======== mainThread ========
 */
void *mainThread(void *arg0)
{
    pthread_t           thread0;
    pthread_attr_t      attrs;
    struct sched_param  priParam;
    int                 retc;
    int                 detachState;

    /* Call driver init functions. */
    Display_init();
    GPIO_init();
    SPI_init();

    /* Configure the LED pins */
    GPIO_setConfig(Board_GPIO_LED0, GPIO_CFG_OUT_STD | GPIO_CFG_OUT_LOW);
    GPIO_setConfig(Board_GPIO_LED1, GPIO_CFG_OUT_STD | GPIO_CFG_OUT_LOW);

    /* Open the display for output */
    display = Display_open(Display_Type_UART, NULL);
    if (display == NULL) {
        /* Failed to open display driver */
        while (1);
    }

    /* Turn on user LED */
    GPIO_write(Board_GPIO_LED0, Board_GPIO_LED_ON);

    Display_printf(display, 0, 0, "Starting the SPI master example");
    Display_printf(display, 0, 0, "This example requires external wires to be "
        "connected to the header pins. Please see the Board.html for details.\n");

    /* Create application threads */
    pthread_attr_init(&attrs);

    detachState = PTHREAD_CREATE_DETACHED;
    /* Set priority and stack size attributes */
    retc = pthread_attr_setdetachstate(&attrs, detachState);
    if (retc != 0) {
        /* pthread_attr_setdetachstate() failed */
        while (1);
    }

    retc |= pthread_attr_setstacksize(&attrs, THREADSTACKSIZE);
    if (retc != 0) {
        /* pthread_attr_setstacksize() failed */
        while (1);
    }

    /* Create master thread */
    priParam.sched_priority = 1;
    pthread_attr_setschedparam(&attrs, &priParam);

    retc = pthread_create(&thread0, &attrs, masterThread, NULL);
    if (retc != 0) {
        /* pthread_create() failed */
        while (1);
    }

    return (NULL);
}

void read_adc_channels()
{
    for (i = 0; i < 2; i++) {
        masterSpi = SPI_open(Board_SPI_MASTER, &spiParams);
        /* Copy message to transmit buffer */
        masterTxBuffer[0] = MASTER_MSG[0];
        transaction.count = 1;
        transaction.txBuf = (void *) masterTxBuffer;
        transaction.rxBuf = (void *) masterRxBuffer;

        GPIO_toggle(Board_GPIO_LED1);

        /* Perform SPI transfer */
        transferOK = SPI_transfer(masterSpi, &transaction);
        if (transferOK) {
            RX_Buffer_1 = masterRxBuffer[0] & 0x0FFF;
        }
        else {
            Display_printf(display, 0, 0, "Unsuccessful master SPI transfer");
        }
        SPI_close(masterSpi);
    }
    //*******************************************************//

    //*******************************************************//
    for (i = 0; i < 2; i++) {
        masterSpi = SPI_open(Board_SPI_MASTER, &spiParams);
        /* Copy message to transmit buffer */
        masterTxBuffer[0] = MASTER_MSG[1];
        transaction.count = 1;
        transaction.txBuf = (void *) masterTxBuffer;
        transaction.rxBuf = (void *) masterRxBuffer;

        GPIO_toggle(Board_GPIO_LED1);

        /* Perform SPI transfer */
        transferOK = SPI_transfer(masterSpi, &transaction);
        if (transferOK) {
            RX_Buffer_2 = masterRxBuffer[0] & 0x0FFF;
        }
        else {
            Display_printf(display, 0, 0, "Unsuccessful master SPI transfer");
        }
        SPI_close(masterSpi);
    }
    //*******************************************************//

    //*******************************************************//
    for (i = 0; i < 2; i++) {
        masterSpi = SPI_open(Board_SPI_MASTER, &spiParams);
        /* Copy message to transmit buffer */
        masterTxBuffer[0] = MASTER_MSG[2];
        transaction.count = 1;
        transaction.txBuf = (void *) masterTxBuffer;
        transaction.rxBuf = (void *) masterRxBuffer;

        GPIO_toggle(Board_GPIO_LED1);

        /* Perform SPI transfer */
        transferOK = SPI_transfer(masterSpi, &transaction);
        if (transferOK) {
            RX_Buffer_3 = masterRxBuffer[0] & 0x0FFF;
        }
        else {
            Display_printf(display, 0, 0, "Unsuccessful master SPI transfer");
        }
        SPI_close(masterSpi);
    }
    //*******************************************************//

    //*******************************************************//
    for (i = 0; i < 2; i++) {
        masterSpi = SPI_open(Board_SPI_MASTER, &spiParams);
        /* Copy message to transmit buffer */
        masterTxBuffer[0] = MASTER_MSG[3];
        transaction.count = 1;
        transaction.txBuf = (void *) masterTxBuffer;
        transaction.rxBuf = (void *) masterRxBuffer;

        GPIO_toggle(Board_GPIO_LED1);

        /* Perform SPI transfer */
        transferOK = SPI_transfer(masterSpi, &transaction);
        if (transferOK) {
            RX_Buffer_4 = masterRxBuffer[0] & 0x0FFF;
        }
        else {
            Display_printf(display, 0, 0, "Unsuccessful master SPI transfer");
        }
        SPI_close(masterSpi);
    }
    //*******************************************************//

    Display_printf(display, 0, 0, "ADC Channels: %lu -- %lu -- %lu -- %lu", RX_Buffer_1, RX_Buffer_2, RX_Buffer_3, RX_Buffer_4);
}
